UNIT FFT;

INTERFACE

CONST
    max_n_samples    = 1024;
    max_n_samples_m1 = 1023;
    max_n_spec       =  512;

TYPE
    complex = RECORD re, im: SINGLE END;
    fourier_signal = ARRAY [0 .. max_n_samples_m1] OF SINGLE;
    fourier_spec   = ARRAY [0 .. max_n_spec]       OF complex;
    fourier_record = RECORD CASE INTEGER OF
        1: (s: fourier_signal);
        2: (a: fourier_spec);
    END;
    db_spec = ARRAY [0..max_n_spec] OF INTEGER;


FUNCTION  TwoTo(val: INTEGER): INTEGER;

PROCEDURE ClearFourierSpec(VAR a: fourier_spec; m: INTEGER);

PROCEDURE ClearFourierSignal(VAR s: fourier_signal; m: INTEGER);

PROCEDURE ComplexFourierTransform(VAR a: fourier_spec; m: INTEGER);

PROCEDURE ComplexInverseTransform(VAR a: fourier_spec; m: INTEGER);

PROCEDURE RealFourierTransform(VAR frec: fourier_record; m: INTEGER);

PROCEDURE ComputedBSpec(

    VAR a: fourier_spec; m: INTEGER; VAR db: db_spec; VAR n: INTEGER
);


IMPLEMENTATION

TYPE long_complex = RECORD re, im: DOUBLE END;


FUNCTION TwoTo(val: INTEGER): INTEGER;

    VAR i, result: INTEGER;

BEGIN
    result := 1;
    FOR i := 1 TO val DO result := result*2;
    TwoTo := result;
END;


PROCEDURE ClearFourierSpec(VAR a: fourier_spec; m: INTEGER);

    VAR i: INTEGER;

BEGIN
    FOR i := 0 TO TwoTo(m) DO BEGIN
        a[i].re := 0;
        a[i].im := 0;
    END;
END;


PROCEDURE ClearFourierSignal(VAR s: fourier_signal; m: INTEGER);

    VAR i: INTEGER;

BEGIN
    FOR i := 0 TO Pred(TwoTo(m)) DO s[i] := 0;
END;


PROCEDURE ComplexFourierTransform(VAR a: fourier_spec; m: INTEGER);

    VAR
        n, nv2, jj, i, j, k, l, lev2v2, lev2, le, wcount: INTEGER;
        temp: complex;
        long_temp, wval, dwval: long_complex;
        dtheta: DOUBLE;


BEGIN
    n := TwoTo(m);

    j := 0;
    nv2 := n DIV 2;
    FOR i := 0 TO n-2 DO BEGIN
        IF i < j THEN BEGIN
            temp := a[i];
            a[i] := a[j];
            a[j] := temp;
        END;
        k := nv2;
        WHILE k <= j DO BEGIN
            j := j - k;
            k := k DIV 2;
        END;
        j := j + k;
    END;

    dtheta := 2.0*Pi;
    le := 1;
    lev2 := 0;

    FOR l := 1 TO m DO BEGIN
        lev2v2 := lev2;
        lev2 := le;
        le := le * 2;

        wval.re := 1.0;
        wval.im := 0.0;
        wcount := 0;
        dtheta := dtheta / 2.0;

        dwval.re :=  cos(dtheta);
        dwval.im := -sin(dtheta);

        jj := 0;

        FOR i := 0 TO Pred(lev2) DO BEGIN
            j := jj;
            WHILE j < n DO BEGIN
                k := j + lev2;
                temp.re := a[k].re*wval.re - a[k].im*wval.im;
                temp.im := a[k].im*wval.re + a[k].re*wval.im;
                a[k].re := a[j].re - temp.re;
                a[k].im := a[j].im - temp.im;
                a[j].re := a[j].re + temp.re;
                a[j].im := a[j].im + temp.im;
                j := j + le;
            END;

            jj := jj + 1;

            wcount := wcount + 1;
            long_temp := wval;
            wval.re := long_temp.re*dwval.re - long_temp.im*dwval.im;
            wval.im := long_temp.re*dwval.im + long_temp.im*dwval.re;
        END;

    END;
END;


PROCEDURE ComplexInverseTransform(VAR a: fourier_spec; m: INTEGER);

    VAR
        i, n: INTEGER;
        fn: SINGLE;

BEGIN
    n := TwoTo(m);
    fn := n;
    FOR i := 0 TO Pred(n) DO BEGIN
        a[i].im := -a[i].im;
    END;
    ComplexFourierTransForm(a,m);
    FOR i := 0 TO Pred(n) DO BEGIN
        a[i].re :=  a[i].re/fn;
        a[i].im := -a[i].im/fn;
    END;
END;


PROCEDURE RealFourierTransform(VAR frec: fourier_record; m: INTEGER);

    VAR nv2, nv4, k: INTEGER;
        long_temp, wval, dwval: long_complex;
        xe, xo, wxo: complex;

BEGIN
    ComplexFourierTransform(frec.a,Pred(m));
    nv2 := TwoTo(Pred(m));
    nv4 := nv2 DIV 2;
    frec.a[nv2].im := 0;
    frec.a[nv2].re := frec.a[0].re - frec.a[0].im;
    frec.a[0].re   := frec.a[0].re + frec.a[0].im;
    frec.a[0].im   := 0;

    wval.re := 1;
    wval.im := 0;
    dwval.re :=  Cos(Pi/nv2);
    dwval.im := -Sin(Pi/nv2);
    FOR k := 1 TO nv4 DO WITH frec DO BEGIN
        long_temp := wval;
        wval.re := long_temp.re*dwval.re - long_temp.im*dwval.im;
        wval.im := long_temp.re*dwval.im + long_temp.im*dwval.re;
        xe.re := 0.5*(a[k].re + a[nv2-k].re);
        xe.im := 0.5*(a[k].im - a[nv2-k].im);
        xo.re := 0.5*(a[k].im + a[nv2-k].im);
        xo.im := 0.5*(a[nv2-k].re - a[k].re);
        wxo.re := xo.re*wval.re - xo.im*wval.im;
        wxo.im := xo.im*wval.re + xo.re*wval.im;
        a[k].re     :=  xe.re + wxo.re;
        a[k].im     :=  xe.im + wxo.im;
        a[nv2-k].re :=  xe.re - wxo.re;
        a[nv2-k].im := -xe.im + wxo.im;
    END;
END;


PROCEDURE ComputedBSpec(

    VAR a: fourier_spec; m: INTEGER; VAR db: db_spec; VAR n: INTEGER
);

    VAR
        i: INTEGER;
        db_scale: DOUBLE;

BEGIN
    db_scale := 10.0/ln(10.0);
    n := TwoTo(Pred(m));
    FOR i := 0 TO n DO WITH a[i] DO
        IF re*re + im*im > 0 THEN
            db[i] :=  Round(100.0*db_scale*Ln(re*re + im*im))
        ELSE
            db[i] := -10000;
END;

END.
